/*
CallStats.java
Copyright (C) 2010  Belledonne Communications, Grenoble, France

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

package org.linphone.core;

import androidx.annotation.Nullable;
import androidx.annotation.NonNull;
import java.lang.StackTraceElement;
import org.linphone.mediastream.Log;

/**
* This object carry various statistic informations regarding the quality of an<br/>
* audio or video stream for a given {@link Call}. <br/>
* <br/>
* To receive these informations periodically and as soon as they are computed,<br/>
* implement the call_stats_updated() callback inside a LinphoneCoreCbs.<br/>
* At any time, the application can access latest computed statistics using {@link Call#getAudioStats}<br/>
* and {@link Call#getVideoStats}. <br/>
*/
public interface CallStats {
    /**
    * Get the bandwidth measurement of the received stream, expressed in kbit/s,<br/>
    * including IP/UDP/RTP headers. <br/>
    * <br/>
    * @return The bandwidth measurement of the received stream in kbit/s. <br/>
    */
    public float getDownloadBandwidth();

    /**
    * Get the estimated bandwidth measurement of the received stream, expressed in<br/>
    * kbit/s, including IP/UDP/RTP headers. <br/>
    * <br/>
    * @return The estimated bandwidth measurement of the received stream in kbit/s. <br/>
    */
    public float getEstimatedDownloadBandwidth();

    /**
    * Get the state of ICE processing. <br/>
    * <br/>
    * @return The {@link IceState} of ICE processing <br/>
    */
    public IceState getIceState();

    /**
    * Get the IP address family of the remote peer. <br/>
    * <br/>
    * @return The IP address family {@link AddressFamily} of the remote peer. <br/>
    */
    public AddressFamily getIpFamilyOfRemote();

    /**
    * Get the jitter buffer size in ms. <br/>
    * <br/>
    * @return The jitter buffer size in ms. <br/>
    */
    public float getJitterBufferSizeMs();

    /**
    * Gets the cumulative number of late packets. <br/>
    * <br/>
    * @return The cumulative number of late packets <br/>
    */
    public int getLatePacketsCumulativeNumber();

    /**
    * Gets the local late rate since last report. <br/>
    * <br/>
    * @return The local late rate <br/>
    */
    public float getLocalLateRate();

    /**
    * Get the local loss rate since last report. <br/>
    * <br/>
    * @return The local loss rate <br/>
    */
    public float getLocalLossRate();

    /**
    * Gets the remote reported interarrival jitter. <br/>
    * <br/>
    * @return The interarrival jitter at last received receiver report <br/>
    */
    public float getReceiverInterarrivalJitter();

    /**
    * Gets the remote reported loss rate since last report. <br/>
    * <br/>
    * @return The receiver loss rate <br/>
    */
    public float getReceiverLossRate();

    /**
    * Get the round trip delay in s. <br/>
    * <br/>
    * @return The round trip delay in s. <br/>
    */
    public float getRoundTripDelay();

    /**
    * Get the bandwidth measurement of the received RTCP, expressed in kbit/s,<br/>
    * including IP/UDP/RTP headers. <br/>
    * <br/>
    * @return The bandwidth measurement of the received RTCP in kbit/s. <br/>
    */
    public float getRtcpDownloadBandwidth();

    /**
    * Get the bandwidth measurement of the sent RTCP, expressed in kbit/s, including<br/>
    * IP/UDP/RTP headers. <br/>
    * <br/>
    * @return The bandwidth measurement of the sent RTCP in kbit/s. <br/>
    */
    public float getRtcpUploadBandwidth();

    /**
    * Gets the local interarrival jitter. <br/>
    * <br/>
    * @return The interarrival jitter at last emitted sender report <br/>
    */
    public float getSenderInterarrivalJitter();

    /**
    * Get the local loss rate since last report. <br/>
    * <br/>
    * @return The sender loss rate <br/>
    */
    public float getSenderLossRate();

    /**
    * Get the type of the stream the stats refer to. <br/>
    * <br/>
    * @return The {@link StreamType} the stats refer to <br/>
    */
    public StreamType getType();

    /**
    * Get the bandwidth measurement of the sent stream, expressed in kbit/s,<br/>
    * including IP/UDP/RTP headers. <br/>
    * <br/>
    * @return The bandwidth measurement of the sent stream in kbit/s. <br/>
    */
    public float getUploadBandwidth();

    /**
    * Get the state of uPnP processing. <br/>
    * <br/>
    * @return The {@link UpnpState} of uPnP processing. <br/>
    */
    public UpnpState getUpnpState();

    /**
      * Sets the object to store in this object user's data
      * @param data the object to store
      */
    public void setUserData(Object data);

    /**
      * Gets the object stored in this object user's data
      * @return the object store if any, null otherwise
      */
    public Object getUserData();

    /**
      * Gets the native pointer used by this class to make native method calls.
      * @return the nativer pointer, as long
      */
    public long getNativePointer();

    public String toString();
}

class CallStatsImpl implements CallStats {

    protected long nativePtr = 0;
    protected transient Object userData = null;
    protected boolean _isConst = false;

    protected CallStatsImpl(long ptr, boolean isConst) {
        nativePtr = ptr;
        _isConst = isConst;
    }

    public long getNativePointer() {
        return nativePtr;
    }


    public boolean isConst() {
        return _isConst;
    }

    private native float getDownloadBandwidth(long nativePtr);
    @Override
    synchronized public float getDownloadBandwidth()  {
        
        
        return getDownloadBandwidth(nativePtr);
    }

    private native float getEstimatedDownloadBandwidth(long nativePtr);
    @Override
    synchronized public float getEstimatedDownloadBandwidth()  {
        
        
        return getEstimatedDownloadBandwidth(nativePtr);
    }

    private native int getIceState(long nativePtr);
    @Override
    synchronized public IceState getIceState()  {
        
        
        return IceState.fromInt(getIceState(nativePtr));
    }

    private native int getIpFamilyOfRemote(long nativePtr);
    @Override
    synchronized public AddressFamily getIpFamilyOfRemote()  {
        
        
        return AddressFamily.fromInt(getIpFamilyOfRemote(nativePtr));
    }

    private native float getJitterBufferSizeMs(long nativePtr);
    @Override
    synchronized public float getJitterBufferSizeMs()  {
        
        
        return getJitterBufferSizeMs(nativePtr);
    }

    private native int getLatePacketsCumulativeNumber(long nativePtr);
    @Override
    synchronized public int getLatePacketsCumulativeNumber()  {
        
        
        return getLatePacketsCumulativeNumber(nativePtr);
    }

    private native float getLocalLateRate(long nativePtr);
    @Override
    synchronized public float getLocalLateRate()  {
        
        
        return getLocalLateRate(nativePtr);
    }

    private native float getLocalLossRate(long nativePtr);
    @Override
    synchronized public float getLocalLossRate()  {
        
        
        return getLocalLossRate(nativePtr);
    }

    private native float getReceiverInterarrivalJitter(long nativePtr);
    @Override
    synchronized public float getReceiverInterarrivalJitter()  {
        
        
        return getReceiverInterarrivalJitter(nativePtr);
    }

    private native float getReceiverLossRate(long nativePtr);
    @Override
    synchronized public float getReceiverLossRate()  {
        
        
        return getReceiverLossRate(nativePtr);
    }

    private native float getRoundTripDelay(long nativePtr);
    @Override
    synchronized public float getRoundTripDelay()  {
        
        
        return getRoundTripDelay(nativePtr);
    }

    private native float getRtcpDownloadBandwidth(long nativePtr);
    @Override
    synchronized public float getRtcpDownloadBandwidth()  {
        
        
        return getRtcpDownloadBandwidth(nativePtr);
    }

    private native float getRtcpUploadBandwidth(long nativePtr);
    @Override
    synchronized public float getRtcpUploadBandwidth()  {
        
        
        return getRtcpUploadBandwidth(nativePtr);
    }

    private native float getSenderInterarrivalJitter(long nativePtr);
    @Override
    synchronized public float getSenderInterarrivalJitter()  {
        
        
        return getSenderInterarrivalJitter(nativePtr);
    }

    private native float getSenderLossRate(long nativePtr);
    @Override
    synchronized public float getSenderLossRate()  {
        
        
        return getSenderLossRate(nativePtr);
    }

    private native int getType(long nativePtr);
    @Override
    synchronized public StreamType getType()  {
        
        
        return StreamType.fromInt(getType(nativePtr));
    }

    private native float getUploadBandwidth(long nativePtr);
    @Override
    synchronized public float getUploadBandwidth()  {
        
        
        return getUploadBandwidth(nativePtr);
    }

    private native int getUpnpState(long nativePtr);
    @Override
    synchronized public UpnpState getUpnpState()  {
        
        
        return UpnpState.fromInt(getUpnpState(nativePtr));
    }

    private native boolean unref(long ptr, boolean isConst);
    protected void finalize() throws Throwable {
        /* Considering the following scenario:
        User scrolls fast in a chat message list that contains a lot of images, 
        so the VM will garbage collect often due to the images taking much space.
        The view itself only has a reference on the EventLogs, not the ChatMessages directly,
        so each call of onBindView will call eventLog.getChatMessage() and the reference will only be temporary.
        It has been observed a race condition can occurs in which the JNI level can obtain a LocalRef from a WeakGlobalRef
        of an object that has been scheduled for destruction and for which the finalize() method will be called shortly after.
        The result is a Java object that has been finalized can have it's methods be called after the finalize() has terminated.*/
        
        /*The workaround we made here is to keep the nativePtr value even after the unref() only if the object is still reffed by the SDK,
        so the few methods calls following the finalize() will still work because we know the native pointer still exists.*/

        /*This solution will work as long as the objects still have a ref at C level, which is the case with ChatMessages because they are hold by the Events.
        Should we have this problem again in a situation where the C no longer holds a reference (resulting in the C object being  destroyed by finalize()), 
        another solution is possible (but more complex to implement).
        This solution would consist in implementing at C level, a queue of belle_sip_object_t that are loosing their last ref. 
        Instead of being deleted immediately by belle_sip_object_unref() calling belle_sip_object_delete(), 
        unref() would place the object onto this "C finalizer queue", the delete() call would then be performed later, 
        by a hook added into the main android looper.
        This solution should work because the actual destruction will be performed from the looper context, 
        so in a place where the local java object causing our troubles no longer exists for sure.
        This solution has however the inconvenience that it can work within an application that uses a single thread to call liblinphone API.*/
        
        if (nativePtr != 0) {
            boolean destroyed = unref(nativePtr, _isConst);
            if (destroyed) {
                nativePtr = 0;
            }
        }
        super.finalize();
    }

    @Override
    public void setUserData(Object data) {
        userData = data;
    }

    @Override
    public Object getUserData() {
        return userData;
    }

    @Override
    public String toString() {
        return "Java object [" + super.toString() + "], native pointer [" + String.format("0x%08x", nativePtr) + "]";
    }
}
